
#ifdef __BORLANDC__
	// Borland stuff.
	#include <windows.h>
	#pragma hdrstop
#endif


// Primary include.
#include "f00Quat.h"

// System includes.
#include <math.h>
#include <assert.h>

// Local includes.
#include "f00PI.h"



// Threshold at which to declare equivalent to 0-length.
#define DEF_Epsilon					0.000001


f00Quat::f00Quat()
{
	reset();
}


f00Quat::f00Quat( const f00Quat& copyFrom )
{
	set( copyFrom );
}


f00Quat::f00Quat( GLfloat x, GLfloat y, GLfloat z )
{
	set( x, y, z );
}


f00Quat::f00Quat( GLfloat angle, GLfloat x, GLfloat y, GLfloat z )
{
	set( angle, x, y, z );
}


f00Quat&
f00Quat::reset()
{
	// The multiplication 'identity' quaternion.  Why not?  :-)
	d_val[ 0 ] = 1.0;
	d_val[ 1 ] = 0.0;
	d_val[ 2 ] = 0.0;
	d_val[ 3 ] = 0.0;

	return *this;
}


f00Quat&
f00Quat::set( const f00Quat& copyFrom )
{
	d_val[ 0 ] = copyFrom.d_val[ 0 ];
	d_val[ 1 ] = copyFrom.d_val[ 1 ];
	d_val[ 2 ] = copyFrom.d_val[ 2 ];
	d_val[ 3 ] = copyFrom.d_val[ 3 ];
	return *this;
}


f00Quat&
f00Quat::set( GLfloat x, GLfloat y, GLfloat z )
{
	f00Quat xQ( x, 1.0, 0.0, 0.0 );
	f00Quat yQ( y, 0.0, 1.0, 0.0 );
	f00Quat zQ( z, 0.0, 0.0, 1.0 );

	set( xQ );
	postMult( yQ );
	postMult( zQ );
	return *this;
}


f00Quat&
f00Quat::set( GLfloat angle, GLfloat x, GLfloat y, GLfloat z )
{
	// Convert to radians.
	angle *= ( M_PI / 180.0 );

	// Normalize the axis.
	GLfloat factor = x * x + y * y + z * z;
	assert( factor != 0 );

	GLfloat scaleBy( 1.0 / sqrt( factor ) );
	x = x * scaleBy;
	y = y * scaleBy;
	z = z * scaleBy;

	// Build a quaternion!
	d_val[ 0 ] = cos( angle / 2.0 );

	GLfloat sinHalfAngle( sin( angle / 2.0 ) );
	d_val[ 1 ] = x * sinHalfAngle;
	d_val[ 2 ] = y * sinHalfAngle;
	d_val[ 3 ] = z * sinHalfAngle;

	return *this;
}


f00Quat&
f00Quat::postMult( const f00Quat& quat )
{
	f00Quat tempQ( *this );
	multAndSet( tempQ, quat );
	return *this;
}


f00Quat&
f00Quat::multAndSet( const f00Quat& quat1, const f00Quat& quat2 )
{
	d_val[ 0 ] =	  quat2.d_val[ 0 ] * quat1.d_val[ 0 ]
						- quat2.d_val[ 1 ] * quat1.d_val[ 1 ]
						- quat2.d_val[ 2 ] * quat1.d_val[ 2 ]
						- quat2.d_val[ 3 ] * quat1.d_val[ 3 ];

	d_val[ 1 ] =	  quat2.d_val[ 0 ] * quat1.d_val[ 1 ]
						+ quat2.d_val[ 1 ] * quat1.d_val[ 0 ]
						+ quat2.d_val[ 2 ] * quat1.d_val[ 3 ]
						- quat2.d_val[ 3 ] * quat1.d_val[ 2 ];

	d_val[ 2 ] =	  quat2.d_val[ 0 ] * quat1.d_val[ 2 ]
						- quat2.d_val[ 1 ] * quat1.d_val[ 3 ]
						+ quat2.d_val[ 2 ] * quat1.d_val[ 0 ]
						+ quat2.d_val[ 3 ] * quat1.d_val[ 1 ];

	d_val[ 3 ] =	  quat2.d_val[ 0 ] * quat1.d_val[ 3 ]
						+ quat2.d_val[ 1 ] * quat1.d_val[ 2 ]
						- quat2.d_val[ 2 ] * quat1.d_val[ 1 ]
						+ quat2.d_val[ 3 ] * quat1.d_val[ 0 ];

	return *this;
}


f00Quat&
f00Quat::normalize()
{
	GLfloat factor =	d_val[ 0 ] * d_val[ 0 ] +
							d_val[ 1 ] * d_val[ 1 ] +
							d_val[ 2 ] * d_val[ 2 ] +
							d_val[ 3 ] * d_val[ 3 ];
	assert( factor != 0.0 );

	if( factor != 1.0 )
	{
		GLfloat scaleBy( 1.0 / sqrt( factor ) );
		d_val[ 0 ] = d_val[ 0 ] * scaleBy;
		d_val[ 1 ] = d_val[ 1 ] * scaleBy;
		d_val[ 2 ] = d_val[ 2 ] * scaleBy;
		d_val[ 3 ] = d_val[ 3 ] * scaleBy;
	}

	return *this;
}


GLfloat*
f00Quat::getMatrix( bool doNormalize )
{
	static GLfloat matrix[ 16 ];
	getMatrix( matrix, doNormalize );
	return matrix;
}


void
f00Quat::getMatrix( GLfloat matrix[ 16 ], bool doNormalize )
{
	if( doNormalize )
		normalize();

	GLfloat w = d_val[ 0 ];
	GLfloat x = d_val[ 1 ];
	GLfloat y = d_val[ 2 ];
	GLfloat z = d_val[ 3 ];

	GLfloat xx = x * x;
	GLfloat yy = y * y;
	GLfloat zz = z * z;

	#define M( x, y )  matrix[ x + y * 4 ]

	M( 0, 0 ) = 1.0 - 2.0 * ( yy + zz );
	M( 1, 0 ) = 2.0 * ( x * y + w * z );
	M( 2, 0 ) = 2.0 * ( x * z - w * y );
	M( 3, 0 ) = 0.0;

	M( 0, 1 ) = 2.0 * ( x * y - w * z );
	M( 1, 1 ) = 1.0 - 2.0 * ( xx + zz );
	M( 2, 1 ) = 2.0 * ( y * z + w * x );
	M( 3, 1 ) = 0.0;

	M( 0, 2 ) = 2.0 * ( x * z + w * y );
	M( 1, 2 ) = 2.0 * ( y * z - w * x );
	M( 2, 2 ) = 1.0 - 2.0 * ( xx + yy );
	M( 3, 2 ) = 0.0;

	M( 0, 3 ) = 0.0;
	M( 1, 3 ) = 0.0;
	M( 2, 3 ) = 0.0;
	M( 3, 3 ) = 1.0;

	#undef M
}


void
f00Quat::getAxisAngle( GLfloat& axisX, GLfloat& axisY, GLfloat& axisZ, GLfloat& rotAngle )
{
	GLfloat lenOfVectorSquared( getSquaredLen() );
	if( lenOfVectorSquared < DEF_Epsilon )
	{
		axisX = 1.0;
		axisY = 0.0;
		axisZ = 0.0;
		rotAngle = 0.0;
	}
	else
	{
		GLfloat invLen( 1.0 / sqrt( lenOfVectorSquared ) );
		axisX = d_val[ 1 ] * invLen;
		axisY = d_val[ 2 ] * invLen;
		axisZ = d_val[ 3 ] * invLen;
		rotAngle = 2.0 * acos( d_val[ 0 ] );
	}
}


void
f00Quat::getLatLong( GLfloat& latAngle, GLfloat& longAngle )
{
	GLfloat dirX;
	GLfloat dirY;
	GLfloat dirZ;
	getDirectionVector( dirX, dirY, dirZ );

	latAngle = asin( dirY ) * ( 180.0 / M_PI );
	if( fabs( dirY ) > ( 1.0 - DEF_Epsilon ) )
	{
		// Pointing straight up or down, the longitude makes no difference.
		longAngle = 0.0;
	}
	else
	{
		if( fabs( dirZ ) < DEF_Epsilon )
			longAngle = ( dirX > 0.0 ? 90.0 : 270.0 );
		else
		{
			if( dirZ >= 0.0 )
				longAngle = atan( dirX / dirZ ) * ( 180.0 / M_PI );
			else
				longAngle = 180.0 + atan( dirX / dirZ ) * ( 180.0 / M_PI );

			if( longAngle < 0.0 )
				longAngle += 360.0;
		}
	}
}


void
f00Quat::getDirectionVector( GLfloat& dirX, GLfloat& dirY, GLfloat& dirZ )
{
	normalize();

	GLfloat w = d_val[ 0 ];
	GLfloat x = d_val[ 1 ];
	GLfloat y = d_val[ 2 ];
	GLfloat z = d_val[ 3 ];

	// Optimized out of the 'getMatrix' function, above.
	dirX = 2.0 * ( x * z - w * y );
	dirY = 2.0 * ( y * z + w * x );
	dirZ = 1.0 - 2.0 * ( x * x + y * y );
}


GLfloat
f00Quat::getSquaredLen() const
{
	return d_val[ 1 ] * d_val[ 1 ] + d_val[ 2 ] * d_val[ 2 ] + d_val[ 3 ] * d_val[ 3 ];
}


void
f00Quat::invert()
{
	d_val[ 0 ] = d_val[ 0 ];
	d_val[ 1 ] = -d_val[ 1 ];
	d_val[ 2 ] = -d_val[ 2 ];
	d_val[ 3 ] = -d_val[ 3 ];
}


bool
f00Quat::isZero() const
{
	return( getSquaredLen() < DEF_Epsilon );
}


f00Point3D
f00Quat::rotatePoint( const f00Point3D& p1, bool doNormalize )
{
	GLfloat* pRotationMatrix( getMatrix( doNormalize ) );
	f00Point3D resultPoint;

	resultPoint.x( pRotationMatrix[  0 ] * p1.x() + pRotationMatrix[  4 ] * p1.y() + pRotationMatrix[  8 ] * p1.z() );
	resultPoint.y( pRotationMatrix[  1 ] * p1.x() + pRotationMatrix[  5 ] * p1.y() + pRotationMatrix[  9 ] * p1.z() );
	resultPoint.z( pRotationMatrix[  2 ] * p1.x() + pRotationMatrix[  6 ] * p1.y() + pRotationMatrix[ 10 ] * p1.z() );

	return resultPoint;
}


GLfloat
f00Quat::dotProduct( const f00Quat& q1 ) const
{
	 return( d_val[ 0 ] * q1.d_val[ 0 ] + d_val[ 1 ] * q1.d_val[ 1 ] + d_val[ 2 ] * q1.d_val[ 2 ] + d_val[ 3 ] * q1.d_val[ 3 ] );
}


f00Quat
f00Quat::operator * ( const f00Quat& q1 ) const
{
	return f00Quat( *this ).postMult( q1 );
}


f00Quat&
f00Quat::operator *= ( const f00Quat& q1 )
{
	return postMult( q1 );
}


f00Quat
f00Quat::operator + ( const f00Quat& q1 ) const
{
	return( f00Quat( *this ) += q1 );
}


f00Quat&
f00Quat::operator += ( const f00Quat& q1 )
{
	d_val[ 0 ] += q1.d_val[ 0 ];
	d_val[ 1 ] += q1.d_val[ 1 ];
	d_val[ 2 ] += q1.d_val[ 2 ];
	d_val[ 3 ] += q1.d_val[ 3 ];
	return *this;
}


f00Quat
f00Quat::operator - ( const f00Quat& q1 ) const
{
	return( f00Quat( *this ) -= q1 );
}


f00Quat&
f00Quat::operator -= ( const f00Quat& q1 )
{
	d_val[ 0 ] -= q1.d_val[ 0 ];
	d_val[ 1 ] -= q1.d_val[ 1 ];
	d_val[ 2 ] -= q1.d_val[ 2 ];
	d_val[ 3 ] -= q1.d_val[ 3 ];
	return *this;
}



