
#ifdef __BORLANDC__
	// Borland stuff.
	#include <windows.h>
	#pragma hdrstop
#endif


// Primary include.
#include "f00Tentacle.h"

// System includes.
#include <math.h>

// Local includes.
#include "f00PI.h"
#include "f00TextureObject.h"



f00Tentacle::f00Tentacle( f00TextureObject* pTexture, int boneCount, int stacksPerBone, int stacksPerCap, int slicesPerStack, float baseRadius, float length, float boneLengthRatio, float boneRadiusRatio )
:	f00BoneSystem(),
	d_boneCount( boneCount ),
	d_stacksPerBone( stacksPerBone ),
	d_stacksPerCap( stacksPerCap ),
	d_slicesPerStack( slicesPerStack ),
	d_pTexture( pTexture ),
	d_isTextured( true )
{
	// Clamps.
	if( d_boneCount < 1 )
		d_boneCount = 1;
	if( d_stacksPerBone < 1 )
		d_stacksPerBone = 1;
	if( d_slicesPerStack < 3 )
		d_slicesPerStack = 3;
	if( d_stacksPerCap < 1 )
		d_stacksPerCap = 1;


	//
	// Build up the non-adjustable things (bone counts, vertex cloud, display elements).
	//

	// Bones.
	BoneInfo* pParentBone( 0 );
	for( int loop = 0; loop < d_boneCount; loop++ )
	{
		BoneInfo* pBI = new BoneInfo;
		pBI->d_pParent = ( loop == 0 ) ? 0 : pParentBone;
		addBone( pBI );
		pParentBone = pBI;
	}

	// Build cubic bias tables.
	std::vector< float > bias1;
	std::vector< float > bias2;
	std::vector< float > bias3;
	for( int step = 0; step <= d_stacksPerBone; step++ )
	{
		float p1( step / (float)d_stacksPerBone );		// Parametric, 0.0 to 1.0
		float p2( p1 * p1 );										// Squared
		bias1.push_back( 0.5 * p2 );
		bias2.push_back( -p2 + p1 + 0.5 );
		bias3.push_back( 0.5 * p2 - p1 + 0.5 );
	}

	// Verts for main body.
	for( int bone = 0; bone < d_boneCount; bone++ )
	{
		BoneInfo* pBI( getBone( bone ) );
		for( int stack = 0; stack < d_stacksPerBone; stack++ )
		{
			for( int slice = 0; slice <= d_slicesPerStack; slice++ )
			{
				VertInfo vI;
				d_vertInfoList.push_back( vI );

				// Link it up.
				GLuint vertIndex( (int)d_vertInfoList.size() - 1 );

				// Previous bone.
				BoneInfo* pPrevBone( getBone( bone - 1 ) );
				if( pPrevBone != 0 )
					pPrevBone->d_linkList.push_back( new LinkInfo( pPrevBone, bias3[ stack ], vertIndex ) );

				// Current bone.
				pBI->d_linkList.push_back( new LinkInfo( pBI, bias2[ stack ], vertIndex ) );

				// Next bone.
				BoneInfo* pNextBone( getBone( bone + 1 ) );
				if( pNextBone != 0 )
					pNextBone->d_linkList.push_back( new LinkInfo( pNextBone, bias1[ stack ], vertIndex ) );
			}
		}
	}

	// Verts for cap.
	BoneInfo* pLastBone( getBone( d_boneCount - 1 ) );
	for( int stack = 0; stack < d_stacksPerCap; stack++ )
	{
		for( int slice = 0; slice <= d_slicesPerStack; slice++ )
		{
			VertInfo vI;
			d_vertInfoList.push_back( vI );

			// Link it to the last bone.
			GLuint vertIndex( (int)d_vertInfoList.size() - 1 );
			pLastBone->d_linkList.push_back( new LinkInfo( pLastBone, 1.0, vertIndex ) );
		}
	}

	// Tip of cap.
	VertInfo vI;
	d_vertInfoList.push_back( vI );
	GLuint vertIndex( (int)d_vertInfoList.size() - 1 );
	pLastBone->d_linkList.push_back( new LinkInfo( pLastBone, 1.0, vertIndex ) );


	// A TRIANGLE_STRIP for each stack in the main body.
	int stackCount( d_stacksPerBone * d_boneCount );
	for( int stack = 0; stack < stackCount; stack++ )
	{
		ElementInfo* pEI = new ElementInfo;
		pEI->d_elemEntryType = GL_TRIANGLE_STRIP;
		int baseIndex( stack * ( d_slicesPerStack + 1 ) );
		for( int slice = 0; slice <= d_slicesPerStack; slice++ )
		{
			pEI->d_elemEntryList.push_back( baseIndex + slice + d_slicesPerStack + 1 );
			pEI->d_elemEntryList.push_back( baseIndex + slice );
		}
		d_elemList.push_back( pEI );
	}

	// A TRIANGLE_STRIP for each stack in the cap (except the last one).
	int capIndex( d_boneCount * d_stacksPerBone * ( d_slicesPerStack + 1 ) );
	for( int stack = 0; stack < ( d_stacksPerCap - 1 ); stack++ )
	{
		ElementInfo* pEI = new ElementInfo;
		pEI->d_elemEntryType = GL_TRIANGLE_STRIP;
		for( int slice = 0; slice <= d_slicesPerStack; slice++ )
		{
			pEI->d_elemEntryList.push_back( capIndex + ( d_slicesPerStack + 1 ) );
			pEI->d_elemEntryList.push_back( capIndex );
			capIndex++;
		}
		d_elemList.push_back( pEI );
	}

	// A TRIANGLE_FAN for the last stack in the cap.
	ElementInfo* pEI = new ElementInfo;
	pEI->d_elemEntryType = GL_TRIANGLE_FAN;
	pEI->d_elemEntryList.push_back( capIndex + ( d_slicesPerStack + 1 ) );	// Center first.
	for( int slice = 0; slice <= d_slicesPerStack; slice++ )
		pEI->d_elemEntryList.push_back( capIndex++ );	// Rim.
	d_elemList.push_back( pEI );


	// Set up the adjustable stuff.
	setParameters( baseRadius, length, boneLengthRatio, boneRadiusRatio );
}


f00Tentacle::~f00Tentacle()
{
	// FIXME Cleanup!
}


void
f00Tentacle::setParameters( float baseRadius, float length, float boneLengthRatio, float boneRadiusRatio )
{
	// Store them.
	d_baseRadius = baseRadius;
	d_length = length;
	d_boneLengthRatio = boneLengthRatio;
	d_boneRadiusRatio = boneRadiusRatio;

	// Clamps.
	if( d_baseRadius < 0.0 )
		d_baseRadius = 0.0;
	if( d_length < 0.0 )
		d_length = 0.0;
	if( d_boneLengthRatio < 0.0 )
		d_boneLengthRatio = 0.0;
	if( d_boneLengthRatio > 1.0 )
		d_boneLengthRatio = 1.0;
	if( d_boneRadiusRatio < 0.0 )
		d_boneRadiusRatio = 0.0;
	if( d_boneRadiusRatio > 1.0 )
		d_boneRadiusRatio = 1.0;

	// Precalculate some settings.
	float boneLengthFirst( d_length / (float)d_boneCount );
	float stackLengthRatio( pow( d_boneLengthRatio, 1.0 / (float)( d_stacksPerBone * d_stacksPerBone ) ) );
	float stackLengthFirstRatio( 1.0 / (float)d_stacksPerBone );
	if( d_boneLengthRatio < 1.0 )
	{
		boneLengthFirst = ( d_length * ( 1.0 - d_boneLengthRatio ) ) / ( 1.0 - pow( d_boneLengthRatio, d_boneCount ) );
		stackLengthFirstRatio = ( ( 1.0 - stackLengthRatio ) ) / ( 1.0 - pow( stackLengthRatio, d_stacksPerBone ) );
	}

	// Set the bone/vertex attributes.
	float curLength( boneLengthFirst );
	float curBoneRadius( d_baseRadius );
	int curIndex( 0 );
	float radiusFactor( pow( d_boneRadiusRatio, ( 1.0 / (float)d_stacksPerBone ) ) );
	float lengthOffset( 0.0 );
	float curRadius( d_baseRadius );
	float curTexCoord( 0.0 );
	for( int bone = 0; bone < d_boneCount; bone++ )
	{
		// Bone attribs.
		BoneInfo* pBI( getBone( bone ) );
		pBI->d_length = curLength;
		pBI->d_baseRadius = curBoneRadius;
		pBI->d_orgBase = ( pBI->d_pParent == 0 ) ? f00Point3D( 0.0, 0.0, 0.0 ) : pBI->d_pParent->d_orgEnd;
		pBI->d_orgEnd = pBI->d_orgBase + f00Point3D( 0.0, 0.0, pBI->d_length );
		curLength *= d_boneLengthRatio;
		curBoneRadius *= d_boneRadiusRatio;

		// Vertex attribs.
		curTexCoord = (float)bone;
		float curBoneLength( pBI->d_length );
		float curStackOffset( lengthOffset );
		float curStackLength( curBoneLength * stackLengthFirstRatio );
		float curTexCoordLength( stackLengthFirstRatio );
		for( int stack = 0; stack < d_stacksPerBone; stack++ )
		{
			float z( curStackOffset );
			for( int slice = 0; slice <= d_slicesPerStack; slice++ )
			{
				float parametric( slice / (float)d_slicesPerStack );
				float theta( M_PI * 2.0 * parametric );
				float x( cos( theta ) );
				float y( sin( theta ) );
				VertInfo& vI( d_vertInfoList[ curIndex++ ] );
				vI.d_vertOrg = f00Point3D( x * curRadius, y * curRadius, z );
				vI.d_texcOrg = f00Point3D( parametric, curTexCoord );
			}
			curRadius *= radiusFactor;
			curStackOffset += curStackLength;
			curTexCoord += curTexCoordLength;
			curTexCoordLength *= stackLengthRatio;
		}
		lengthOffset += curBoneLength;
	}

	// Add the cap.
	float capRadius( curRadius );
	float capTexCoord( curTexCoord );
	float capTexCoordIncrement( d_boneLengthRatio / (float)d_stacksPerCap );
	for( int stack = 0; stack < d_stacksPerCap; stack++ )
	{
		float capParamteric( stack / (float)d_stacksPerCap );
		float capTheta( capParamteric * M_PI / 2.0 );
		float z( d_length + capRadius * sin( capTheta ) );
		float capScalar( capRadius * cos( capTheta ) );
		for( int slice = 0; slice <= d_slicesPerStack; slice++ )
		{
			float parametric( slice / (float)d_slicesPerStack );
			float theta( M_PI * 2.0 * parametric );
			float x( capScalar * cos( theta ) );
			float y( capScalar * sin( theta ) );
			VertInfo& vI( d_vertInfoList[ curIndex++ ] );
			vI.d_vertOrg = f00Point3D( x, y, z );
			vI.d_texcOrg = f00Point3D( parametric, capTexCoord );
		}
		capTexCoord += capTexCoordIncrement;
	}

	// Finally, the tip vertex.
	VertInfo& vI( d_vertInfoList[ curIndex++ ] );
	vI.d_vertOrg = f00Point3D( 0.0, 0.0, d_length + capRadius );
	vI.d_texcOrg = f00Point3D( 0.5, capTexCoord );
}


void
f00Tentacle::renderObject( const f00Point3D& viewPos, const f00Quat& viewDir, const f00TimeBase& timeBase )
{
	// Smack the base object.
	f00BoneSystem::renderObject( viewPos, viewDir, timeBase );


	//
	// Clone from master lists.
	//

	std::vector< f00Point3D >				d_texcList;
	std::vector< f00Point3D >				d_normList;
	std::vector< f00Point3D >				d_vertList;
	std::vector< float >						d_vertWeightList;
	for( int vert = 0; vert < (int)d_vertInfoList.size(); vert++ )
	{
		VertInfo& vI( d_vertInfoList[ vert ] );
		d_texcList.push_back( vI.d_texcOrg );
		d_normList.push_back( f00Point3D( 0.0, 0.0, 0.0 ) );
		d_vertList.push_back( vI.d_vertOrg );
		d_vertWeightList.push_back( 0.0 );
	}


	//
	// Move the verts.
	//

	// Walk the bone's links, accumulating vertex weight factors.
	for( int bone = 0; bone < countBones(); bone++ )
	{
		BoneInfo* pBI( getBone( bone ) );
		for( int link = 0; link < (int)pBI->d_linkList.size(); link++ )
		{
			LinkInfo& lI( *pBI->d_linkList[ link ] );
			d_vertWeightList[ lI.d_vertIndex ] += lI.d_weight;
		}
	}

	// Walk the bones, accumulating xforms.
	f00Quat curOrient;
	f00Point3D curEnd( 0.0, 0.0, 0.0 );
	for( int bone = 0; bone < countBones(); bone++ )
	{
		BoneInfo* pBI( getBone( bone ) );
		curOrient.postMult( pBI->d_orientation );
		f00Point3D localEnd( curOrient.rotatePoint( f00Point3D( 0.0, 0.0, pBI->d_length ) ) );

		// Walk the links, xforming verts.
		for( int link = 0; link < (int)pBI->d_linkList.size(); link++ )
		{
			LinkInfo& lI( *pBI->d_linkList[ link ] );
			f00Point3D localDelta( d_vertInfoList[ lI.d_vertIndex ].d_vertOrg - pBI->d_orgBase );
			f00Point3D vertDelta( curOrient.rotatePoint( localDelta, false ) - localDelta );
			d_vertList[ lI.d_vertIndex ] += ( vertDelta + ( curEnd - pBI->d_orgBase ) ) * ( lI.d_weight / d_vertWeightList[ lI.d_vertIndex ] );
		}
		curEnd += localEnd;
	}


	//
	// Update the normals.
	//

	// Walk all the elements, cross each one to get a face normal, add it to each vertex.
	for( int elem = 0; elem < (int)d_elemList.size(); elem++ )
	{
		ElementInfo& eI( *d_elemList[ elem ] );
		int elemEntryCount( (int)eI.d_elemEntryList.size() );
		switch( eI.d_elemEntryType )
		{
			case GL_POINTS:
			case GL_LINES:
			case GL_LINE_STRIP:
			case GL_LINE_LOOP:
				{
					// Normals are meaningless - ignore.
				}
				break;

			case GL_TRIANGLES:
				{
					elemEntryCount -= ( elemEntryCount % 3 );
					for( int loop = 0; loop < elemEntryCount; loop += 3 )
					{
						// A normal to the face may be found by crossing the first and second edges.
						GLuint elemEntryA( eI.d_elemEntryList[ loop + 0 ] );
						GLuint elemEntryB( eI.d_elemEntryList[ loop + 1 ] );
						GLuint elemEntryC( eI.d_elemEntryList[ loop + 2 ] );
						f00Point3D vect1( d_vertList[ elemEntryC ] - d_vertList[ elemEntryB ] );
						f00Point3D vect2( d_vertList[ elemEntryA ] - d_vertList[ elemEntryB ] );
						f00Point3D normal( f00Point3D::crossProduct( vect1, vect2 ) );
						d_normList[ eI.d_elemEntryList[ loop + 0 ] ] += normal;
						d_normList[ eI.d_elemEntryList[ loop + 1 ] ] += normal;
						d_normList[ eI.d_elemEntryList[ loop + 2 ] ] += normal;
					}
				}
				break;

			case GL_QUADS:
				{
					elemEntryCount -= ( elemEntryCount % 4 );
					for( int loop = 0; loop < elemEntryCount; loop += 4 )
					{
						// A normal to the face may be found by crossing the first and second edges.
						// FIXME  Degenerate case: initially-colinear
						GLuint elemEntryA( eI.d_elemEntryList[ loop + 0 ] );
						GLuint elemEntryB( eI.d_elemEntryList[ loop + 1 ] );
						GLuint elemEntryC( eI.d_elemEntryList[ loop + 2 ] );
						f00Point3D vect1( d_vertList[ elemEntryC ] - d_vertList[ elemEntryB ] );
						f00Point3D vect2( d_vertList[ elemEntryA ] - d_vertList[ elemEntryB ] );
						f00Point3D normal( f00Point3D::crossProduct( vect1, vect2 ) );
						d_normList[ eI.d_elemEntryList[ loop + 0 ] ] += normal;
						d_normList[ eI.d_elemEntryList[ loop + 1 ] ] += normal;
						d_normList[ eI.d_elemEntryList[ loop + 2 ] ] += normal;
						d_normList[ eI.d_elemEntryList[ loop + 3 ] ] += normal;
					}
				}
				break;

			case GL_POLYGON:
				{
					// A normal to the face may be found by crossing the first and second edges.
					// FIXME  Degenerate case: initially-colinear
					GLuint elemEntryA( eI.d_elemEntryList[ 0 ] );
					GLuint elemEntryB( eI.d_elemEntryList[ 1 ] );
					GLuint elemEntryC( eI.d_elemEntryList[ 2 ] );
					f00Point3D vect1( d_vertList[ elemEntryC ] - d_vertList[ elemEntryB ] );
					f00Point3D vect2( d_vertList[ elemEntryA ] - d_vertList[ elemEntryB ] );
					f00Point3D normal( f00Point3D::crossProduct( vect1, vect2 ) );
					for( int elemEntry = 0; elemEntry < (int)eI.d_elemEntryList.size(); elemEntry++ )
						d_normList[ eI.d_elemEntryList[ elemEntry ] ] += normal;
				}
				break;

			case GL_TRIANGLE_FAN:
				{
					if( elemEntryCount >= 3 )
					{
						GLuint elemEntryA( eI.d_elemEntryList[ 0 ] );
						elemEntryCount -= 2;
						bool evenTriCount( elemEntryCount % 2 == 0 );
						for( int loop = 0; loop < elemEntryCount; loop += 2 )
						{
							GLuint elemEntryB( eI.d_elemEntryList[ loop + 1 ] );
							GLuint elemEntryC( eI.d_elemEntryList[ loop + 2 ] );
							f00Point3D vect1( d_vertList[ elemEntryA ] - d_vertList[ elemEntryB ] );
							f00Point3D vect2( d_vertList[ elemEntryC ] - d_vertList[ elemEntryB ] );
							f00Point3D normal( f00Point3D::crossProduct( vect2, vect1 ) );
							d_normList[ eI.d_elemEntryList[ loop + 0 ] ] += normal;
							d_normList[ eI.d_elemEntryList[ loop + 1 ] ] += normal;
							d_normList[ eI.d_elemEntryList[ loop + 2 ] ] += normal;

							if( evenTriCount )
							{
								// Do another one, too.
								GLuint elemEntryD( eI.d_elemEntryList[ loop + 3 ] );
								f00Point3D vect3( d_vertList[ elemEntryC ] - d_vertList[ elemEntryD ] );
								f00Point3D normal( f00Point3D::crossProduct( vect2, vect3 ) );
								d_normList[ eI.d_elemEntryList[ loop + 1 ] ] += normal;
								d_normList[ eI.d_elemEntryList[ loop + 2 ] ] += normal;
								d_normList[ eI.d_elemEntryList[ loop + 3 ] ] += normal;
							}
						}
					}
				}
				break;

			case GL_TRIANGLE_STRIP:
				{
					if( elemEntryCount >= 3 )
					{
						bool evenTriCount( elemEntryCount % 2 == 0 );
						elemEntryCount -= 2;
						for( int loop = 0; loop < elemEntryCount; loop += 2 )
						{
							GLuint elemEntryA( eI.d_elemEntryList[ loop + 0 ] );
							GLuint elemEntryB( eI.d_elemEntryList[ loop + 1 ] );
							GLuint elemEntryC( eI.d_elemEntryList[ loop + 2 ] );
							f00Point3D vect1( d_vertList[ elemEntryA ] - d_vertList[ elemEntryB ] );
							f00Point3D vect2( d_vertList[ elemEntryC ] - d_vertList[ elemEntryB ] );
							f00Point3D normal( f00Point3D::crossProduct( vect2, vect1 ) );
							d_normList[ eI.d_elemEntryList[ loop + 0 ] ] += normal;
							d_normList[ eI.d_elemEntryList[ loop + 1 ] ] += normal;
							d_normList[ eI.d_elemEntryList[ loop + 2 ] ] += normal;

							if( evenTriCount )
							{
								// Do another one, too.
								GLuint elemEntryD( eI.d_elemEntryList[ loop + 3 ] );
								f00Point3D vect3( d_vertList[ elemEntryC ] - d_vertList[ elemEntryD ] );
								f00Point3D normal( f00Point3D::crossProduct( vect2, vect3 ) );
								d_normList[ eI.d_elemEntryList[ loop + 1 ] ] += normal;
								d_normList[ eI.d_elemEntryList[ loop + 2 ] ] += normal;
								d_normList[ eI.d_elemEntryList[ loop + 3 ] ] += normal;
							}
						}
					}
				}
				break;

			case GL_QUAD_STRIP:
				{
					if( elemEntryCount >= 4 )
					{
						elemEntryCount -= 2;
						for( int loop = 0; loop < elemEntryCount; loop += 2 )
						{
							GLuint elemEntryA( eI.d_elemEntryList[ loop + 0 ] );
							GLuint elemEntryB( eI.d_elemEntryList[ loop + 1 ] );
							GLuint elemEntryC( eI.d_elemEntryList[ loop + 2 ] );
							f00Point3D vect1( d_vertList[ elemEntryA ] - d_vertList[ elemEntryB ] );
							f00Point3D vect2( d_vertList[ elemEntryC ] - d_vertList[ elemEntryB ] );
							f00Point3D normal( f00Point3D::crossProduct( vect2, vect1 ) );
							d_normList[ eI.d_elemEntryList[ loop + 0 ] ] += normal;
							d_normList[ eI.d_elemEntryList[ loop + 1 ] ] += normal;
							d_normList[ eI.d_elemEntryList[ loop + 2 ] ] += normal;
							d_normList[ eI.d_elemEntryList[ loop + 3 ] ] += normal;
						}
					}
				}
				break;

			default:
				break;

		}
	}

	// Walk the lists, finalizing values.
	for( int vert = 0; vert < (int)d_vertList.size(); vert++ )
		d_normList[ vert ].normalize();


	//
	// Draw stuff now.
	//

	// TexCoords.
	if( d_isTextured )
	{
		d_pTexture->select();
		glEnable( GL_TEXTURE_2D );

		glEnableClientState( GL_TEXTURE_COORD_ARRAY );
		glTexCoordPointer( 2, GL_FLOAT, sizeof( f00Point3D ), d_texcList[ 0 ].getArray() );
	}
	else
		glDisable( GL_TEXTURE_2D );

	// Normals.
	glEnableClientState( GL_NORMAL_ARRAY );
	glNormalPointer( GL_FLOAT, sizeof( f00Point3D ), d_normList[ 0 ].getArray() );

	// Verts.
	glEnableClientState( GL_VERTEX_ARRAY );
	glVertexPointer( 3, GL_FLOAT, sizeof( f00Point3D ), d_vertList[ 0 ].getArray() );

	// Draw!
	for( int loop = 0; loop < (int)d_elemList.size(); loop++ )
	{
		ElementInfo& eI( *d_elemList[ loop ] );
		glDrawElements( eI.d_elemEntryType, eI.d_elemEntryList.size(), GL_UNSIGNED_INT, &( eI.d_elemEntryList[ 0 ] ) );
	}

	// Cleanup.
	glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );
	if( d_isTextured )
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );

}



