#ifndef f00TextureObjectH
#define f00TextureObjectH


// Local includes.
#include "f00ImageLoader.h"


//----------------------------------------------------------------------------------
//
// NOTE: These classes are designed to work with OpenGL 1.1 or higher (due to the
// use of texture 'objects' (namespaces)....
//
//----------------------------------------------------------------------------------


//
// f00RAMTexture
// - Used to manage an in-memory array of GLubyte entries (RGBA).
//

class f00RAMTexture
{

	public:

		// Ownership moves in!
		f00RAMTexture( GLubyte* pImageRGBA, int width, int height );

		f00RAMTexture( const f00RAMTexture& copyFrom );

		~f00RAMTexture();


	public:

		void
		clear();


		int
		incrementRefCount()			{ return ++d_refCount; }

		int
		decrementRefCount()			{ return --d_refCount; }


		// Ownership moves in!
		bool
		set( GLubyte* pImageRGBA, int width, int height );


		// Ownership stays in!
		GLubyte*
		getImageRGBA() const			{ return d_pImageRGBA; }

		int
		width() const					{ return d_width; }

		int
		height() const					{ return d_height; }


	private:

		GLubyte*				d_pImageRGBA;
		int					d_width;
		int					d_height;

		int					d_refCount;

};



//
// f00GLTexture
// - Encapsulation of an OpenGL 'texture object'.
//

class f00GLTexture
{

	public:

		f00GLTexture();

private:
		// Not Yet Implemented - Do Not Use!
		f00GLTexture( const f00GLTexture& copyFrom );
public:

		f00GLTexture( f00RAMTexture& rRAMTex, GLint magFilter = GL_NONE, GLint minFilter = GL_NONE, GLint wrapS = GL_REPEAT, GLint wrapT = GL_REPEAT );

		virtual
		~f00GLTexture();


	public:

		bool
		loadTexture( f00RAMTexture& rRAMTex, GLint magFilter = GL_NONE, GLint minFilter = GL_NONE, GLint wrapS = GL_REPEAT, GLint wrapT = GL_REPEAT );

		bool
		selectTexture() const;


		GLuint
		textureID() const										{ return d_textureID; }


		static GLint
		magFilterDefault()									{ return d_magFilterDefault; }

		static void
		magFilterDefault( GLint magFilterDefault )	{ d_magFilterDefault = magFilterDefault; }


		static GLint
		minFilterDefault()									{ return d_minFilterDefault; }

		static void
		minFilterDefault( GLint minFilterDefault )	{ d_minFilterDefault = minFilterDefault; }


		static bool
		isMIPMap( GLint minFilter )
		{
			return(	( minFilter == GL_NEAREST_MIPMAP_NEAREST ) ||
						( minFilter == GL_LINEAR_MIPMAP_NEAREST ) ||
						( minFilter == GL_NEAREST_MIPMAP_LINEAR ) ||
						( minFilter == GL_LINEAR_MIPMAP_LINEAR )		);
		}


	private:

		GLuint					d_textureID;

		static GLint			d_magFilterDefault;
		static GLint			d_minFilterDefault;

};



//
// f00TextureObject
// - Manages certain aspects of a texture, including:
//   - A RAM-buffer allowing for soft-updatable textures.
//   - Shared access to a RAM texture (ie: for multi-context painting).
//   - One OpenGL 'texture object'.
//

class f00TextureObject
{

	public:

		// Ownership moves in (if specified).
		f00TextureObject( f00RAMTexture* pRAMTexture = 0, GLint magFilter = GL_NONE, GLint minFilter = GL_NONE, GLint wrapS = GL_REPEAT, GLint wrapT = GL_REPEAT );

		f00TextureObject( const f00TextureObject& copyFrom, bool shareRAMTexture = true, GLint magFilter = GL_NONE, GLint minFilter = GL_NONE );

		f00TextureObject( f00ImageLoader& anImageLoader, bool extractImagePointer = true, GLint magFilter = GL_NONE, GLint minFilter = GL_NONE, GLint wrapS = GL_REPEAT, GLint wrapT = GL_REPEAT );

		virtual
		~f00TextureObject();


	public:

		void
		setTextureQuality( GLint magFilter = GL_NONE, GLint minFilter = GL_NONE );

		void
		setWrapMode( GLint wrapS, GLint wrapT );

		bool
		loadRAMIntoGL( GLint magFilter = GL_NONE, GLint minFilter = GL_NONE );

		bool
		select() const;


		// Ownership stays in!
		f00RAMTexture*
		getRAMTexture() const				{ return d_pRAMTexture; }

		GLuint
		textureID() const						{ return d_textureID; }


	private:

		f00GLTexture			d_myGLTexture;
		f00RAMTexture*			d_pRAMTexture;

		GLint						d_magFilter;
		GLint						d_minFilter;

		GLint						d_wrapS;
		GLint						d_wrapT;

		GLuint					d_textureID;

};


#endif

